import _socket
import math
import socket

ENCODING = 'utf-8'
PREFIX_LENGTH_SIZED = 10  # default length for a message's length prefix with '(send|recv)_sized'
DEFAULT_CHUNK_SIZE = 500  # default size of a chunk in 'send_chopped'
PREFIX_LENGTH_CHOPPED = 30  # default length for a message's length prefix with '(send|recv)_chopped'
HOME_ADDRESS = '127.0.0.1'
# this will be used as the port that the server will listen to
# and the client will connect to:
RANDOM_PORT = 6969


class Socket(socket.socket):

    def __init__(self, *args, **kwargs):
        super(Socket, self).__init__(*args, **kwargs)

    @classmethod
    def copy(cls, sock: socket.socket):
        """turn a socket to a Socket"""
        with sock:
            fd = _socket.dup(sock.fileno())
            copy = cls(sock.family, sock.type, sock.proto, fileno=fd)
            copy.settimeout(sock.gettimeout())
        return copy

    def send_sized(self, message: [str, bytes]):
        """
        send the given message, prefixed with it's size.
        message size is up to 4 digits because it is enough.
        """
        if (type(message) is str) or (type(message) is bytes):
            self.sendall(to_bytes(str(len(message)).ljust(PREFIX_LENGTH_SIZED)) +
                         to_bytes(message))
        else:
            raise TypeError('message has to be either string or bytes')

    def recv_sized(self) -> bytes:
        """recieves a message, prefixed with its size of PREFIX_LENGTH_SIZED digits."""
        size = int(to_str(self.recv(PREFIX_LENGTH_SIZED)))
        return self.recv(size)

    def send_chopped(self, message: [str, bytes],
                     chunk_size=DEFAULT_CHUNK_SIZE,
                     prefix_length=PREFIX_LENGTH_CHOPPED):
        """send the message in chunks."""
        chunk_num = math.floor(len(message) / chunk_size)
        last_chunk_len = len(message) - chunk_num * chunk_size
        prefix = f'{chunk_num},{chunk_size},{last_chunk_len}' \
            .ljust(prefix_length)
        self.sendall(to_bytes(prefix) + to_bytes(message))

    def recv_chopped(self, prefix_length=PREFIX_LENGTH_CHOPPED) -> bytes:
        """receive a message prefixed with:
        '{chunk_num},{chunk_size},{last_chunk_len}' in chunks."""
        prefix = self.recv(prefix_length)
        chunk_num, chunk_size, last_chunk_len = \
            tuple(prefix.split(b','))
        message = [
            self.recv(int(chunk_size)) for _ in range(int(chunk_num))
        ]
        message.append(self.recv(int(last_chunk_len)))
        return b''.join(message)


def to_bytes(data: [str, bytes]) -> bytes:
    """turns data into bytes unless it is already bytes"""
    if type(data) is str:
        return bytes(data, ENCODING)
    else:
        return data


def to_str(data: [str, bytes]) -> str:
    """turns data into str unless it is already str"""
    if type(data) is bytes:
        return str(data)[2:-1].replace(r'\n', '\n')
    else:
        return data


def to_Socket(soc: [socket.socket, Socket]) -> Socket:
    """turns soc into Socket unless it is already Socket"""
    if type(soc) is socket.socket:
        return Socket.copy(soc)
    else:
        return soc
